'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
__doc__ = """
The ``pm4py.ocel`` module contains the object-centric process mining features offered in ``pm4py``.
"""

from typing import List, Dict, Collection, Any, Optional, Set, Tuple

import pandas as pd

from pm4py.objects.ocel.obj import OCEL
from pm4py.util import constants, pandas_utils
import sys
import random


def ocel_get_object_types(ocel: OCEL) -> List[str]:
    """
    Returns the list of object types contained in the object-centric event log
    (e.g., ["order", "item", "delivery"]).

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: List of object types.
    :rtype: List[str]

    .. code-block:: python3

        import pm4py

        object_types = pm4py.ocel_get_object_types(ocel)
    """
    return pandas_utils.format_unique(ocel.objects[ocel.object_type_column].unique())


def ocel_get_attribute_names(ocel: OCEL) -> List[str]:
    """
    Returns the list of attributes at the event and object levels of an object-centric event log
    (e.g., ["cost", "amount", "name"]).

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: List of attribute names.
    :rtype: List[str]

    .. code-block:: python3

        import pm4py

        attribute_names = pm4py.ocel_get_attribute_names(ocel)
    """
    from pm4py.objects.ocel.util import attributes_names
    return attributes_names.get_attribute_names(ocel)


def ocel_flattening(ocel: OCEL, object_type: str) -> pd.DataFrame:
    """
    Flattens the object-centric event log to a traditional event log based on a chosen object type.
    In the flattened log, the objects of the specified type are treated as cases, and each case
    contains the set of events related to that object.
    The flattened log follows the XES notations for case identifier, activity, and timestamp. Specifically:
    - "case:concept:name" is used for the case ID.
    - "concept:name" is used for the activity.
    - "time:timestamp" is used for the timestamp.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param object_type: The object type to use as cases.
    :type object_type: str
    :return: Flattened traditional event log.
    :rtype: pd.DataFrame

    .. code-block:: python3

        import pm4py

        event_log = pm4py.ocel_flattening(ocel, 'items')
    """
    from pm4py.objects.ocel.util import flattening
    return flattening.flatten(ocel, object_type)


def ocel_object_type_activities(ocel: OCEL) -> Dict[str, Collection[str]]:
    """
    Returns the set of activities performed for each object type.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Dictionary mapping object types to their associated activities.
    :rtype: Dict[str, Collection[str]]

    .. code-block:: python3

        import pm4py

        ot_activities = pm4py.ocel_object_type_activities(ocel)
    """
    from pm4py.statistics.ocel import ot_activities

    return ot_activities.get_object_type_activities(ocel)


def ocel_objects_ot_count(ocel: OCEL) -> Dict[str, Dict[str, int]]:
    """
    Returns the count of related objects per type for each event.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Nested dictionary mapping events to object types and their counts.
    :rtype: Dict[str, Dict[str, int]]

    .. code-block:: python3

        import pm4py

        objects_ot_count = pm4py.ocel_objects_ot_count(ocel)
    """
    from pm4py.statistics.ocel import objects_ot_count

    return objects_ot_count.get_objects_ot_count(ocel)


def ocel_temporal_summary(ocel: OCEL) -> pd.DataFrame:
    """
    Returns the temporal summary of an object-centric event log.
    The temporal summary aggregates all events that occur at the same timestamp
    and reports the list of activities and involved objects.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Temporal summary DataFrame.
    :rtype: pd.DataFrame

    .. code-block:: python3

        import pm4py

        temporal_summary = pm4py.ocel_temporal_summary(ocel)
    """
    gdf = ocel.relations.groupby(ocel.event_timestamp)
    act_comb = gdf[ocel.event_activity].agg(list).to_frame()
    obj_comb = gdf[ocel.object_id_column].agg(list).to_frame()
    temporal_summary = act_comb.join(obj_comb).reset_index()
    return temporal_summary


def ocel_objects_summary(ocel: OCEL) -> pd.DataFrame:
    """
    Returns the objects summary of an object-centric event log.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Objects summary DataFrame containing lifecycle information and interacting objects.
    :rtype: pd.DataFrame

    .. code-block:: python3

        import pm4py

        objects_summary = pm4py.ocel_objects_summary(ocel)
    """
    gdf = ocel.relations.groupby(ocel.object_id_column)
    act_comb = (
        gdf[ocel.event_activity]
        .agg(list)
        .to_frame()
        .rename(columns={ocel.event_activity: "activities_lifecycle"})
    )
    lif_start_tim = (
        gdf[ocel.event_timestamp]
        .min()
        .to_frame()
        .rename(columns={ocel.event_timestamp: "lifecycle_start"})
    )
    lif_end_tim = (
        gdf[ocel.event_timestamp]
        .max()
        .to_frame()
        .rename(columns={ocel.event_timestamp: "lifecycle_end"})
    )
    objects_summary = act_comb.join(lif_start_tim)
    objects_summary = objects_summary.join(lif_end_tim)
    objects_summary = objects_summary.reset_index()
    objects_summary["lifecycle_duration"] = pandas_utils.get_total_seconds(
        objects_summary["lifecycle_end"] - objects_summary["lifecycle_start"]
    )
    ev_rel_obj = ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column].agg(list).to_dict()
    objects_ids = pandas_utils.format_unique(ocel.objects[ocel.object_id_column].unique())
    graph = {o: set() for o in objects_ids}
    for ev in ev_rel_obj:
        rel_obj = ev_rel_obj[ev]
        for o1 in rel_obj:
            for o2 in rel_obj:
                if o1 != o2:
                    graph[o1].add(o2)
    objects_summary["interacting_objects"] = objects_summary[ocel.object_id_column].map(graph)
    return objects_summary


def ocel_objects_interactions_summary(ocel: OCEL) -> pd.DataFrame:
    """
    Returns the objects interactions summary of an object-centric event log.
    The summary includes a row for every combination of (event, related object, other related object).
    Properties such as the activity of the event and the object types of the two related objects are included.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Objects interactions summary DataFrame.
    :rtype: pd.DataFrame

    .. code-block:: python3

        import pm4py

        interactions_summary = pm4py.ocel_objects_interactions_summary(ocel)
    """
    obj_types = ocel.objects.groupby(ocel.object_id_column)[ocel.object_type_column].first().to_dict()
    eve_activities = ocel.events.groupby(ocel.event_id_column)[ocel.event_activity].first().to_dict()
    ev_rel_obj = ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column].agg(list).to_dict()
    stream = []
    for ev in ev_rel_obj:
        rel_obj = ev_rel_obj[ev]
        for o1 in rel_obj:
            for o2 in rel_obj:
                if o1 != o2:
                    stream.append(
                        {
                            ocel.event_id_column: ev,
                            ocel.event_activity: eve_activities[ev],
                            ocel.object_id_column: o1,
                            ocel.object_type_column: obj_types[o1],
                            f"{ocel.object_id_column}_2": o2,
                            f"{ocel.object_type_column}_2": obj_types[o2],
                        }
                    )

    return pandas_utils.instantiate_dataframe(stream)


def discover_ocdfg(ocel: OCEL, business_hours: bool = False, business_hour_slots: Optional[List[Tuple[int, int]]] = constants.DEFAULT_BUSINESS_HOUR_SLOTS) -> Dict[str, Any]:
    """
    Discovers an Object-Centric Directly-Follows Graph (OC-DFG) from an object-centric event log.

    Object-centric directly-follows multigraphs are a composition of directly-follows graphs for each object type.
    These graphs can be annotated with different metrics considering the entities of an object-centric event log
    (i.e., events, unique objects, total objects).

    Reference paper:
    Berti, Alessandro, and Wil van der Aalst. "Extracting multiple viewpoint models from relational databases." 
    Data-Driven Process Discovery and Analysis. Springer, Cham, 2018. 24-51.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param business_hours: Enable the usage of business hours if set to True.
    :type business_hours: bool
    :param business_hour_slots: Work schedule of the company, provided as a list of tuples where each tuple 
                                 represents one time slot of business hours. Each tuple consists of a start 
                                 and an end time given in seconds since week start, e.g., 
                                 [(25200, 61200), (9072, 43200), (46800, 61200)] meaning that business hours 
                                 are Mondays 07:00 - 17:00, Tuesdays 02:32 - 12:00, and Wednesdays 13:00 - 17:00.
    :type business_hour_slots: Optional[List[Tuple[int, int]]]
    :return: OC-DFG discovery result.
    :rtype: Dict[str, Any]

    .. code-block:: python3

        import pm4py

        ocdfg = pm4py.discover_ocdfg(ocel)
    """
    parameters = {
        "business_hours": business_hours,
        "business_hour_slots": business_hour_slots,
    }
    from pm4py.algo.discovery.ocel.ocdfg import algorithm as ocdfg_discovery
    return ocdfg_discovery.apply(ocel, parameters=parameters)


def discover_oc_petri_net(ocel: OCEL, inductive_miner_variant: str = "im", diagnostics_with_tbr: bool = False) -> Dict[str, Any]:
    """
    Discovers an object-centric Petri net from the provided object-centric event log.

    Reference paper: van der Aalst, Wil MP, and Alessandro Berti. 
    "Discovering object-centric Petri nets." Fundamenta Informaticae 175.1-4 (2020): 1-40.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param inductive_miner_variant: Variant of the inductive miner to use ("im" for traditional; "imd" for the faster inductive miner directly-follows).
    :type inductive_miner_variant: str
    :param diagnostics_with_tbr: Enable the computation of diagnostics using token-based replay if set to True.
    :type diagnostics_with_tbr: bool
    :return: Discovered object-centric Petri net.
    :rtype: Dict[str, Any]

    .. code-block:: python3

        import pm4py

        ocpn = pm4py.discover_oc_petri_net(ocel)
    """
    from pm4py.algo.discovery.ocel.ocpn import algorithm as ocpn_discovery
    parameters = {
        "inductive_miner_variant": inductive_miner_variant,
        "diagnostics_with_token_based_replay": diagnostics_with_tbr,
    }

    return ocpn_discovery.apply(ocel, parameters=parameters)


def discover_objects_graph(ocel: OCEL, graph_type: str = "object_interaction") -> Set[Tuple[str, str]]:
    """
    Discovers an object graph from the provided object-centric event log.

    Available graph types:
    - "object_interaction"
    - "object_descendants"
    - "object_inheritance"
    - "object_cobirth"
    - "object_codeath"

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param graph_type: Type of graph to consider.
                       Options include "object_interaction", "object_descendants", 
                       "object_inheritance", "object_cobirth", "object_codeath".
    :type graph_type: str
    :return: Discovered object graph as a set of tuples.
    :rtype: Set[Tuple[str, str]]

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        obj_graph = pm4py.discover_objects_graph(ocel, graph_type='object_interaction')
    """
    if graph_type == "object_interaction":
        from pm4py.algo.transformation.ocel.graphs import object_interaction_graph
        return object_interaction_graph.apply(ocel)
    elif graph_type == "object_descendants":
        from pm4py.algo.transformation.ocel.graphs import object_descendants_graph
        return object_descendants_graph.apply(ocel)
    elif graph_type == "object_inheritance":
        from pm4py.algo.transformation.ocel.graphs import object_inheritance_graph
        return object_inheritance_graph.apply(ocel)
    elif graph_type == "object_cobirth":
        from pm4py.algo.transformation.ocel.graphs import object_cobirth_graph
        return object_cobirth_graph.apply(ocel)
    elif graph_type == "object_codeath":
        from pm4py.algo.transformation.ocel.graphs import object_codeath_graph
        return object_codeath_graph.apply(ocel)
    else:
        raise ValueError(f"Unsupported graph_type: {graph_type}")


def ocel_o2o_enrichment(ocel: OCEL, included_graphs: Optional[Collection[str]] = None) -> OCEL:
    """
    Enriches the OCEL with information inferred from graph computations by inserting them into the O2O relations.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param included_graphs: Types of graphs to include, provided as a list or set of strings. 
                            Options include "object_interaction_graph", "object_descendants_graph", 
                            "object_inheritance_graph", "object_cobirth_graph", "object_codeath_graph".
    :type included_graphs: Optional[Collection[str]]
    :return: Enriched object-centric event log.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_o2o_enrichment(ocel)
        print(ocel.o2o)
    """
    from pm4py.algo.transformation.ocel.graphs import ocel20_computation
    return ocel20_computation.apply(ocel, parameters={"included_graphs": included_graphs})


def ocel_e2o_lifecycle_enrichment(ocel: OCEL) -> OCEL:
    """
    Enriches the OCEL with lifecycle-based information, indicating when an object is created, terminated, 
    or has other types of relations, by updating the E2O relations.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Enriched object-centric event log with lifecycle information.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_e2o_lifecycle_enrichment(ocel)
        print(ocel.relations)
    """
    from pm4py.objects.ocel.util import e2o_qualification
    ocel = e2o_qualification.apply(ocel, "termination")
    ocel = e2o_qualification.apply(ocel, "creation")
    ocel = e2o_qualification.apply(ocel, "other")
    return ocel


def sample_ocel_objects(ocel: OCEL, num_objects: int) -> OCEL:
    """
    Returns a sampled object-centric event log containing a random subset of objects.
    Only events related to at least one of the sampled objects are included in the returned log.
    Note that this sampling may disrupt the relationships between different objects.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param num_objects: Number of objects to include in the sampled event log.
    :type num_objects: int
    :return: Sampled object-centric event log.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        sampled_ocel = pm4py.sample_ocel_objects(ocel, 50)  # Keeps only 50 random objects
    """
    from pm4py.objects.ocel.util import sampling
    return sampling.sample_ocel_objects(ocel, parameters={"num_entities": num_objects})


def sample_ocel_connected_components(
    ocel: OCEL,
    connected_components: int = 1,
    max_num_events_per_cc: int = sys.maxsize,
    max_num_objects_per_cc: int = sys.maxsize,
    max_num_e2o_relations_per_cc: int = sys.maxsize
) -> OCEL:
    """
    Returns a sampled object-centric event log containing a specified number of connected components.
    Users can also set maximum limits on the number of events, objects, and E2O relations per connected component.

    Reference paper:
    Adams, Jan Niklas, et al. "Defining cases and variants for object-centric event data." 
    2022 4th International Conference on Process Mining (ICPM). IEEE, 2022.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param connected_components: Number of connected components to include in the sampled event log.
    :type connected_components: int
    :param max_num_events_per_cc: Maximum number of events allowed per connected component (default: sys.maxsize).
    :type max_num_events_per_cc: int
    :param max_num_objects_per_cc: Maximum number of objects allowed per connected component (default: sys.maxsize).
    :type max_num_objects_per_cc: int
    :param max_num_e2o_relations_per_cc: Maximum number of event-to-object relationships allowed per connected component (default: sys.maxsize).
    :type max_num_e2o_relations_per_cc: int
    :return: Sampled object-centric event log containing the specified connected components.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        sampled_ocel = pm4py.sample_ocel_connected_components(ocel, 5)  # Keeps only 5 connected components
    """
    from pm4py.algo.transformation.ocel.split_ocel import algorithm
    ocel_splits = algorithm.apply(ocel, variant=algorithm.Variants.CONNECTED_COMPONENTS)
    events = None
    objects = None
    relations = None
    ocel_splits = [
        x for x in ocel_splits
        if len(x.events) <= max_num_events_per_cc
        and len(x.objects) <= max_num_objects_per_cc
        and len(x.relations) <= max_num_e2o_relations_per_cc
    ]

    if len(ocel_splits) > 0:
        ocel_splits = random.sample(ocel_splits, min(connected_components, len(ocel_splits)))

    for cc in ocel_splits:
        if events is None:
            events = cc.events
            objects = cc.objects
            relations = cc.relations
        else:
            events = pandas_utils.concat([events, cc.events])
            objects = pandas_utils.concat([objects, cc.objects])
            relations = pandas_utils.concat([relations, cc.relations])

    return OCEL(events, objects, relations)


def ocel_drop_duplicates(ocel: OCEL) -> OCEL:
    """
    Removes duplicate relations between events and objects that occur at the same time,
    have the same activity, and are linked to the same object identifier.
    This effectively cleans the OCEL by eliminating duplicate events.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Cleaned object-centric event log without duplicate relations.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_drop_duplicates(ocel)
    """
    from pm4py.objects.ocel.util import filtering_utils
    ocel.relations = ocel.relations.drop_duplicates(
        subset=[ocel.event_activity, ocel.event_timestamp, ocel.object_id_column]
    )
    ocel = filtering_utils.propagate_relations_filtering(ocel)
    return ocel


def ocel_merge_duplicates(ocel: OCEL, have_common_object: Optional[bool] = False) -> OCEL:
    """
    Merges events in the OCEL that have the same activity and timestamp. Optionally, ensures that
    the events being merged share a common object.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param have_common_object: If set to True, only merges events that share a common object. Defaults to False.
    :type have_common_object: Optional[bool]
    :return: Object-centric event log with merged duplicate events.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_merge_duplicates(ocel)
    """
    import copy
    import uuid
    relations = copy.copy(ocel.relations)
    if have_common_object:
        relations["@@groupn"] = relations.groupby(
            [ocel.object_id_column, ocel.event_activity, ocel.event_timestamp]
        ).ngroup()
    else:
        relations["@@groupn"] = relations.groupby(
            [ocel.event_activity, ocel.event_timestamp]
        ).ngroup()

    group_size = relations["@@groupn"].value_counts().to_dict()
    relations["@@groupsize"] = relations["@@groupn"].map(group_size)
    relations = relations.sort_values(["@@groupsize", "@@groupn"], ascending=False)
    val_corr = {
        x: str(uuid.uuid4()) for x in pandas_utils.format_unique(relations["@@groupn"].unique())
    }
    relations = relations.groupby(ocel.event_id_column).first()["@@groupn"].to_dict()
    relations = {x: val_corr[y] for x, y in relations.items()}

    ocel.events[ocel.event_id_column] = ocel.events[ocel.event_id_column].map(relations)
    ocel.relations[ocel.event_id_column] = ocel.relations[ocel.event_id_column].map(relations)

    ocel.events = ocel.events.drop_duplicates(subset=[ocel.event_id_column])
    ocel.relations = ocel.relations.drop_duplicates(subset=[ocel.event_id_column, ocel.object_id_column])

    return ocel


def ocel_sort_by_additional_column(
    ocel: OCEL, additional_column: str, primary_column: str = "ocel:timestamp"
) -> OCEL:
    """
    Sorts the OCEL based on the primary timestamp column and an additional column to determine
    the order of events occurring at the same timestamp.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param additional_column: Additional column to use for sorting.
    :type additional_column: str
    :param primary_column: Primary column to use for sorting (default: "ocel:timestamp"). 
                           Typically the timestamp column.
    :type primary_column: str
    :return: Sorted object-centric event log.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_sort_by_additional_column(ocel, 'ordering')
    """
    ocel.events = pandas_utils.insert_index(ocel.events, "@@index", reset_index=False, copy_dataframe=False)
    ocel.events = ocel.events.sort_values([primary_column, additional_column, "@@index"])
    del ocel.events["@@index"]
    ocel.events = ocel.events.reset_index(drop=True)
    return ocel


def ocel_add_index_based_timedelta(ocel: OCEL) -> OCEL:
    """
    Adds a small time delta to the timestamp column based on the event index to ensure the correct ordering
    of events within any object-centric process mining solution.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :return: Object-centric event log with index-based time deltas added.
    :rtype: OCEL

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_add_index_based_timedelta(ocel)
    """
    from datetime import timedelta
    eids = ocel.events[ocel.event_id_column].to_numpy().tolist()
    eids = {eids[i]: timedelta(milliseconds=i) for i in range(len(eids))}
    ocel.events["@@timedelta"] = ocel.events[ocel.event_id_column].map(eids)
    ocel.relations["@@timedelta"] = ocel.relations[ocel.event_id_column].map(eids)
    ocel.events[ocel.event_timestamp] = ocel.events[ocel.event_timestamp] + ocel.events["@@timedelta"]
    ocel.relations[ocel.event_timestamp] = ocel.relations[ocel.event_timestamp] + ocel.relations["@@timedelta"]
    del ocel.events["@@timedelta"]
    del ocel.relations["@@timedelta"]
    return ocel


def cluster_equivalent_ocel(
    ocel: OCEL, object_type: str, max_objs: int = sys.maxsize
) -> Dict[str, Collection[OCEL]]:
    """
    Clusters the object-centric event log based on the 'executions' of a single object type.
    Equivalent 'executions' are grouped together in the output dictionary.

    :param ocel: Object-centric event log.
    :type ocel: OCEL
    :param object_type: Reference object type for clustering.
    :type object_type: str
    :param max_objs: Maximum number of objects (of the specified object type) to include per cluster.
                    Defaults to sys.maxsize.
    :type max_objs: int
    :return: Dictionary mapping cluster descriptions to collections of equivalent OCELs.
    :rtype: Dict[str, Collection[OCEL]]

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        clusters = pm4py.cluster_equivalent_ocel(ocel, "order")
    """
    from pm4py.algo.transformation.ocel.split_ocel import algorithm as split_ocel_algorithm
    from pm4py.objects.ocel.util import rename_objs_ot_tim_lex
    from pm4py.algo.transformation.ocel.description import algorithm as ocel_description

    lst_ocels = split_ocel_algorithm.apply(
        ocel,
        variant=split_ocel_algorithm.Variants.ANCESTORS_DESCENDANTS,
        parameters={"object_type": object_type, "max_objs": max_objs},
    )
    ret = {}
    for index, oc in enumerate(lst_ocels):
        oc_ren = rename_objs_ot_tim_lex.apply(oc)
        descr = ocel_description.apply(oc_ren, parameters={"include_timestamps": False})
        if descr not in ret:
            ret[descr] = []
        ret[descr].append(oc)
    return ret



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.bpmn.obj import BPMN
from typing import Optional, Dict, Any

from copy import copy
import tempfile


def apply(bpmn_graph: BPMN, parameters: Optional[Dict[Any, Any]] = None) -> BPMN:
    """
    Layouts the BPMN graphviz using directly the information about node positioning
    and edges waypoints provided in the SVG obtained from Graphviz.

    Parameters
    -----------------
    bpmn_graph
        BPMN graph
    parameters
        Optional parameters of the method

    Returns
    ----------------
    layouted_bpmn
        Layouted BPMN
    """
    if parameters is None:
        parameters = {}

    from pm4py.visualization.bpmn.variants import classic as bpmn_visualizer
    from pm4py.visualization.common import svg_pos_parser

    layout = bpmn_graph.get_layout()

    filename_svg = tempfile.NamedTemporaryFile(suffix='.svg')
    filename_svg.close()

    vis_parameters = copy(parameters)
    vis_parameters["format"] = "svg"
    vis_parameters["include_name_in_events"] = False
    vis_parameters["endpoints_shape"] = "box"

    gviz = bpmn_visualizer.apply(bpmn_graph, parameters=vis_parameters)
    bpmn_visualizer.save(gviz, filename_svg.name)

    #print(filename_svg.name)

    nodes_p, edges_p = svg_pos_parser.apply(filename_svg.name)

    for node in list(bpmn_graph.get_nodes()):
        node_id = str(id(node))
        if node_id in nodes_p:
            node_info = nodes_p[node_id]
            if node_info["polygon"] is not None:
                min_x = min(x[0] for x in node_info["polygon"])
                max_x = max(x[0] for x in node_info["polygon"])
                min_y = min(x[1] for x in node_info["polygon"])
                max_y = max(x[1] for x in node_info["polygon"])

                width = max_x - min_x
                height = max_y - min_y

                layout.get(node).set_width(width)
                layout.get(node).set_height(height)
                layout.get(node).set_x(min_x)
                layout.get(node).set_y(min_y)

    for flow in list(bpmn_graph.get_flows()):
        flow_id = (str(id(flow.source)), str(id(flow.target)))
        if flow_id in edges_p:
            flow_info = edges_p[flow_id]
            if flow_info["waypoints"] is not None:
                flow.del_waypoints()

                for wayp in flow_info["waypoints"]:
                    flow.add_waypoint(wayp)

    return bpmn_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from enum import Enum
from typing import Optional, List, Dict, Any, Tuple, Union

import numpy as np

from pm4py.objects.log.obj import EventLog
from pm4py.util import exec_utils
from pm4py.objects.conversion.log import converter


class Parameters(Enum):
    STR_EVENT_ATTRIBUTES = "str_ev_attr"
    NUM_EVENT_ATTRIBUTES = "num_ev_attr"
    FEATURE_NAMES = "feature_names"
    MIN_NUM_DIFF_STR_VALUES = "min_num_diff_str_values"
    MAX_NUM_DIFF_STR_VALUES = "max_num_diff_str_values"


def extract_all_ev_features_names_from_log(log: EventLog, str_ev_attr: List[str], num_ev_attr: List[str],
                                           parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> List[str]:
    """
    Extracts the feature names from an event log.

    Parameters
    ---------------
    log
        Event log
    str_ev_attr
        (if provided) list of string event attributes to consider in extracting the feature names
    num_ev_attr
        (if provided) list of integer event attributes to consider in extracting the feature names
    parameters
        Parameters, including:
            - MIN_NUM_DIFF_STR_VALUES => minimum number of distinct values to include an attribute as feature(s)
            - MAX_NUM_DIFF_STR_VALUES => maximum number of distinct values to include an attribute as feature(s)

    Returns
    ----------------
    feature_names
        List of feature names
    """
    if parameters is None:
        parameters = {}

    min_num_diff_str_values = exec_utils.get_param_value(Parameters.MIN_NUM_DIFF_STR_VALUES, parameters, 2)
    max_num_diff_str_values = exec_utils.get_param_value(Parameters.MAX_NUM_DIFF_STR_VALUES, parameters, 500)

    str_features = {}
    num_features = Counter()
    count_events = 0
    for trace in log:
        for event in trace:
            count_events += 1
            for attr_name in event:
                attr_value = event[attr_name]
                if isinstance(attr_value, str) and (str_ev_attr is None or attr_name in str_ev_attr):
                    if attr_name not in str_features:
                        str_features[attr_name] = set()
                    str_features[attr_name].add("event:" + attr_name + "@" + attr_value)
                elif isinstance(attr_value, int) or isinstance(attr_value, float):
                    if num_ev_attr is None or attr_name in num_ev_attr:
                        num_features["event:" + attr_name] += 1
    num_features = list({x for x, y in num_features.items() if y == count_events})
    str_features = list({z for x, y in str_features.items() for z in y if
                         min_num_diff_str_values <= len(y) <= max_num_diff_str_values})

    feature_names = str_features + num_features
    feature_names = sorted(feature_names)

    return feature_names


def extract_features(log: EventLog, feature_names: List[str], parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Tuple[
    Any, List[str]]:
    """
    Extracts the matrix of the features from an event log

    Parameters
    ---------------
    log
        Event log
    feature_names
        Features to consider (in the given order)

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    v1 = max(len(trace) for trace in log)
    v2 = len(feature_names)
    data = np.zeros((len(log), v1, v2), dtype=np.float32)

    for i1, trace in enumerate(log):
        for i2, event in enumerate(trace):
            str_features = set()
            num_features = {}
            for attr_name in event:
                attr_value = event[attr_name]
                if isinstance(attr_value, str):
                    str_features.add("event:" + attr_name + "@" + attr_value)
                elif isinstance(attr_value, int) or isinstance(attr_value, float):
                    num_features["event:" + attr_name] = float(attr_value)
            for attr in str_features:
                if attr in feature_names:
                    data[i1, i2, feature_names.index(attr)] = 1.0
            for attr in num_features:
                if attr in feature_names:
                    data[i1, i2, feature_names.index(attr)] = float(num_features[attr])

    return data, feature_names


def apply(log: EventLog, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Tuple[Any, List[str]]:
    """
    Extracts all the features for the traces of an event log (each trace becomes a vector of vectors, where each
    event has its own vector)

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
            - STR_EVENT_ATTRIBUTES => string event attributes to consider in the features extraction
            - NUM_EVENT_ATTRIBUTES => numeric event attributes to consider in the features extraction
            - FEATURE_NAMES => features to consider (in the given order)

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    if parameters is None:
        parameters = {}

    str_ev_attr = exec_utils.get_param_value(Parameters.STR_EVENT_ATTRIBUTES, parameters, None)
    num_ev_attr = exec_utils.get_param_value(Parameters.NUM_EVENT_ATTRIBUTES, parameters, None)
    feature_names = exec_utils.get_param_value(Parameters.FEATURE_NAMES, parameters, None)

    log = converter.apply(log, variant=converter.Variants.TO_EVENT_LOG, parameters=parameters)

    if feature_names is None:
        feature_names = extract_all_ev_features_names_from_log(log, str_ev_attr, num_ev_attr, parameters=parameters)

    return extract_features(log, feature_names, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.objects.conversion.log.variants import to_event_stream, to_event_log, to_data_frame, to_nx


class Variants(Enum):
    TO_EVENT_LOG = to_event_log
    TO_EVENT_STREAM = to_event_stream
    TO_DATA_FRAME = to_data_frame
    TO_NX = to_nx


TO_EVENT_LOG = Variants.TO_EVENT_LOG
TO_EVENT_STREAM = Variants.TO_EVENT_STREAM
TO_DATA_FRAME = Variants.TO_DATA_FRAME


def apply(log, parameters=None, variant=None):
    if variant is None:
        variant = Variants.TO_EVENT_LOG
    return variant.value.apply(log, parameters=parameters)



import runpy
from os.path import dirname, join
from pathlib import Path
from setuptools import setup, find_packages


# Import only the metadata of the pm4py to use in the setup. We cannot import it directly because
# then we need to import packages that are about to be installed by the setup itself.
meta_path = Path(__file__).parent.absolute() / "pm4py" / "meta.py"
meta = runpy.run_path(str(meta_path))


def read_file(filename):
    with open(join(dirname(__file__), filename)) as f:
        return f.read()


setup(
    name=meta['__name__'],
    version=meta['__version__'],
    description=meta['__doc__'].strip(),
    long_description=read_file('README.md'),
    author=meta['__author__'],
    author_email=meta['__author_email__'],
    py_modules=['pm4py'],
    include_package_data=True,
    packages=[x for x in find_packages() if x.startswith("pm4py")],
    url='https://processintelligence.solutions/',
    license='AGPL 3.0',
    install_requires=read_file("requirements.txt").split("\n"),
    project_urls={
        'Documentation': 'https://processintelligence.solutions/pm4py/',
        'Source': 'https://github.com/pm4py/pm4py-source',
        'Tracker': 'https://github.com/pm4py/pm4py-source/issues',
    }
)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
from enum import Enum
from typing import Any, Optional, Dict, Union

from pm4py.objects.conversion.log import converter
from pm4py.objects.log.obj import EventLog
from pm4py.util import constants, xes_constants, exec_utils


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    MIN_REP = "min_rep"
    MAX_REP = "max_rep"


def apply(log: EventLog, value: Any, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> EventLog:
    """
    Filters the trace of the log where the given attribute value is repeated
    (in a range of repetitions that is specified by the user)

    Parameters
    ----------------
    log
        Event log
    value
        Value that is investigated
    parameters
        Parameters of the filter, including:
        - Parameters.ATTRIBUTE_KEY => the attribute key
        - Parameters.MIN_REP => minimum number of repetitions
        - Parameters.MAX_REP => maximum number of repetitions

    Returns
    ----------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    log = converter.apply(log, variant=converter.Variants.TO_EVENT_LOG, parameters=parameters)

    attribute_key = exec_utils.get_param_value(Parameters.ATTRIBUTE_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    min_rep = exec_utils.get_param_value(Parameters.MIN_REP, parameters, 2)
    max_rep = exec_utils.get_param_value(Parameters.MAX_REP, parameters, sys.maxsize)

    filtered_log = EventLog(list(), attributes=log.attributes, extensions=log.extensions, classifiers=log.classifiers,
                            omni_present=log.omni_present, properties=log.properties)

    for trace in log:
        rep = 0
        for event in trace:
            if attribute_key in event and event[attribute_key] == value:
                rep += 1
        if min_rep <= rep <= max_rep:
            filtered_log.append(trace)

    return filtered_log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel import constants as ocel_constants
import pandas as pd


class Parameters(Enum):
    OCEL_TYPE_PREFIX = ocel_constants.PARAM_OBJECT_TYPE_PREFIX_EXTENDED


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Creates an iterator over the events of an object-centric event log

    Parameters
    ----------------
    ocel
        OCEL
    parameters
        Parameters of the method, including:
        - Parameters.OCEL_TYPE_PREFIX => the prefix of the object types in the OCEL (default: ocel:type)

    Returns
    ----------------
    yielded event
        The events of the OCEL, one by one.
    """
    if parameters is None:
        parameters = {}

    ot_prefix = exec_utils.get_param_value(Parameters.OCEL_TYPE_PREFIX, parameters,
                                           ocel_constants.DEFAULT_OBJECT_TYPE_PREFIX_EXTENDED)

    ext_table = ocel.get_extended_table(ot_prefix)

    for k, ev in ext_table.iterrows():
        yield {x: y for x, y in dict(ev).items() if isinstance(y, list) or not pd.isna(y)}



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from math import sqrt

from pm4py.algo.conformance.tokenreplay import algorithm as token_replay
from enum import Enum
from pm4py.util import constants
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def get_generalization(petri_net, aligned_traces):
    trans_occ_map = Counter()
    for trace in aligned_traces:
        for trans in trace["activated_transitions"]:
            trans_occ_map[trans] += 1
    inv_sq_occ_sum = 0.0
    for trans in trans_occ_map:
        this_term = 1.0 / sqrt(trans_occ_map[trans])
        inv_sq_occ_sum = inv_sq_occ_sum + this_term
    for trans in petri_net.transitions:
        if trans not in trans_occ_map:
            inv_sq_occ_sum = inv_sq_occ_sum + 1
    generalization = 1.0
    if len(petri_net.transitions) > 0:
        generalization = 1.0 - inv_sq_occ_sum / float(len(petri_net.transitions))
    return generalization


def apply(log: Union[EventLog, pd.DataFrame], petri_net: PetriNet, initial_marking: Marking, final_marking: Marking, parameters: Optional[Dict[Union[str, Parameters], Any]] = None):
    if parameters is None:
        parameters = {}

    aligned_traces = token_replay.apply(log, petri_net, initial_marking, final_marking, parameters=parameters)

    return get_generalization(petri_net, aligned_traces)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import shutil
import os
from pm4py.visualization.common import dot_util, html


def save(gviz, output_file_path, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    format = os.path.splitext(output_file_path)[1][1:].lower()
    is_dot_installed = dot_util.check_dot_installed()

    if format.startswith("html"):
        html.save(gviz, output_file_path, parameters=parameters)
    else:
        render = gviz.render(cleanup=True)
        shutil.copyfile(render, output_file_path)
    """elif not is_dot_installed:
        raise Exception("impossible to save formats different from HTML without the Graphviz binary")"""




'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from graphviz import Digraph
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
from pm4py.visualization.ocel.interleavings.variants import graphviz
import pandas as pd


class Variants(Enum):
    GRAPHVIZ = graphviz


def apply(dataframe1: pd.DataFrame, dataframe2: pd.DataFrame, interleavings: pd.DataFrame, variant=Variants.GRAPHVIZ,
          parameters: Optional[Dict[Any, Any]] = None) -> Digraph:
    """
    Visualizes the interleavings discovered between two different processes.
    We suppose to provide both event logs, and the discovered interleavings.
    The visualization includes the DFG of both processes, along with the arcs discovered between them.
    Both frequency and performance visualization are available.

    Parameters
    --------------------
    dataframe1
        Dataframe of the first process
    dataframe2
        Dataframe of the second process
    interleavings
        Interleavings between the two considered processes
    variant
        Variant of the visualizer to apply, possible values: Variants.GRAPHVIZ
    parameters
        Variant-specific parameters

    Returns
    ----------------
    digraph
        Graphviz Digraph
    """
    return exec_utils.get_variant(variant).apply(dataframe1, dataframe2, interleavings, parameters=parameters)


def save(gviz: Digraph, output_file_path: str, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)


def view(gviz: Digraph, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz: Digraph, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """

    return gview.matplotlib_view(gviz, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from pm4py.algo.analysis.woflan.graphs import utility as helper
from pm4py.util import nx_utils


def construct_tree(net, initial_marking):
    """
    Construct a restricted coverability marking.
    For more information, see the thesis "Verification of WF-nets", 4.3.
    :param net:
    :param initial_marking:
    :return:
    """
    initial_marking = helper.convert_marking(net, initial_marking)
    firing_dict = helper.split_incidence_matrix(helper.compute_incidence_matrix(net), net)
    req_dict = helper.compute_firing_requirement(net)
    look_up_indices = {}
    j = 0
    coverability_graph = nx_utils.DiGraph()
    coverability_graph.add_node(j, marking=initial_marking)
    look_up_indices[np.array2string(initial_marking)] = j

    j += 1
    new_arc = True
    while new_arc:
        new_arc = False
        nodes = list(coverability_graph.nodes).copy()
        while len(nodes) > 0:
            m = nodes.pop()
            if not np.inf in coverability_graph.nodes[m]['marking']:
                possible_markings = helper.enabled_markings(firing_dict, req_dict,
                                                            coverability_graph.nodes[m]['marking'])
                m2 = None
                if len(possible_markings) > 0:
                    for marking in possible_markings:
                        # check for m1 + since we want to construct a tree, we do not want that a marking is already in a graph since it is going to have an arc
                        if np.array2string(marking[0]) not in look_up_indices:
                            if check_if_transition_unique(m, coverability_graph, marking[1]):
                                m2 = marking
                                new_arc = True
                                break
                if new_arc:
                    break
        if new_arc:
            lplaces = sorted(list(net.places), key=lambda x: x.name)
            m3 = np.zeros(len(lplaces))
            for place in lplaces:
                if check_for_smaller_marking(m2, coverability_graph, lplaces.index(place), m, look_up_indices):
                    m3[lplaces.index(place)] = np.inf
                else:
                    m3[lplaces.index(place)] = m2[0][lplaces.index(place)]
            coverability_graph.add_node(j, marking=m3)
            coverability_graph.add_edge(m, j, transition=m2[1])
            look_up_indices[np.array2string(m3)] = j
            j += 1
    return coverability_graph


def check_if_transition_unique(marking, graph, transition):
    for edge in graph.out_edges(marking):
        if graph[edge[0]][edge[1]]['transition'] == transition:
            return False
    return True


def check_for_smaller_marking(marking, coverability_graph, index, current_node, look_up_indices):
    for node in coverability_graph.nodes:
        if all(np.less_equal(coverability_graph.nodes[node]['marking'], marking[0])):
            if coverability_graph.nodes[node]['marking'][index] < marking[0][index]:
                if nx_utils.has_path(coverability_graph,
                               look_up_indices[np.array2string(coverability_graph.nodes[node]['marking'])],
                               current_node):
                    return True
    return False



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import List, Dict


class Role(object):
    activities: List[str]
    originator_importance = Dict[str, float]

    def __init__(self, activities: List[str], originator_importance: Dict[str, float]):
        self.activities = activities
        self.originator_importance = originator_importance

    def __repr__(self):
        return "Activities: " + str(self.activities) + " Originators importance " + str(self.originator_importance)

    def __str__(self):
        return "Activities: " + str(self.activities) + " Originators importance " + str(self.originator_importance)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
__doc__ = """
The ``pm4py.filtering`` module contains the filtering features offered in ``pm4py``.
"""

from typing import Union, Set, List, Tuple, Collection, Any, Dict, Optional
from collections import Counter

import pandas as pd

from pm4py.objects.log.obj import EventLog
from pm4py.util import constants, xes_constants, pandas_utils, nx_utils
import warnings
from pm4py.util.pandas_utils import check_is_pandas_dataframe, check_pandas_dataframe_columns
from pm4py.utils import get_properties, __event_log_deprecation_warning
from pm4py.objects.ocel.obj import OCEL
import datetime


def filter_log_relative_occurrence_event_attribute(
    log: Union[EventLog, pd.DataFrame],
    min_relative_stake: float,
    attribute_key: str = xes_constants.DEFAULT_NAME_KEY,
    level: str = "cases",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the event log, keeping only the events that have an attribute value which occurs:
    - in at least the specified (min_relative_stake) percentage of events when level="events",
    - in at least the specified (min_relative_stake) percentage of cases when level="cases".

    :param log: Event log or Pandas DataFrame.
    :param min_relative_stake: Minimum percentage of cases (expressed as a number between 0 and 1) in which the attribute should occur.
    :param attribute_key: The attribute to filter.
    :param level: The level of the filter (if level="events", then events; if level="cases", then cases).
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_log_relative_occurrence_event_attribute(
            dataframe,
            0.5,
            attribute_key='concept:name',
            level='cases',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, timestamp_key=timestamp_key, case_id_key=case_id_key, activity_key=attribute_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.attributes import attributes_filter
        parameters[attributes_filter.Parameters.ATTRIBUTE_KEY] = attribute_key
        parameters[attributes_filter.Parameters.KEEP_ONCE_PER_CASE] = True if level == "cases" else False
        return attributes_filter.filter_df_relative_occurrence_event_attribute(log, min_relative_stake, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.attributes import attributes_filter
        parameters[attributes_filter.Parameters.ATTRIBUTE_KEY] = attribute_key
        parameters[attributes_filter.Parameters.KEEP_ONCE_PER_CASE] = True if level == "cases" else False
        return attributes_filter.filter_log_relative_occurrence_event_attribute(log, min_relative_stake, parameters=parameters)


def filter_start_activities(
    log: Union[EventLog, pd.DataFrame],
    activities: Union[Set[str], List[str]],
    retain: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters cases that have a start activity in the provided list.

    :param log: Event log or Pandas DataFrame.
    :param activities: Collection of start activities.
    :param retain: If True, retains the traces containing the given start activities; if False, drops the traces.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_start_activities(
            dataframe,
            ['Act. A'],
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.start_activities import start_activities_filter
        parameters[start_activities_filter.Parameters.POSITIVE] = retain
        return start_activities_filter.apply(log, activities, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.start_activities import start_activities_filter
        parameters[start_activities_filter.Parameters.POSITIVE] = retain
        return start_activities_filter.apply(log, activities, parameters=parameters)


def filter_end_activities(
    log: Union[EventLog, pd.DataFrame],
    activities: Union[Set[str], List[str]],
    retain: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters cases that have an end activity in the provided list.

    :param log: Event log or Pandas DataFrame.
    :param activities: Collection of end activities.
    :param retain: If True, retains the traces containing the given end activities; if False, drops the traces.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_end_activities(
            dataframe,
            ['Act. Z'],
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.end_activities import end_activities_filter
        parameters[end_activities_filter.Parameters.POSITIVE] = retain
        return end_activities_filter.apply(log, activities, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.end_activities import end_activities_filter
        parameters[end_activities_filter.Parameters.POSITIVE] = retain
        return end_activities_filter.apply(log, activities, parameters=parameters)


def filter_event_attribute_values(
    log: Union[EventLog, pd.DataFrame],
    attribute_key: str,
    values: Union[Set[str], List[str]],
    level: str = "case",
    retain: bool = True,
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters a log object based on the values of a specified event attribute.

    :param log: Event log or Pandas DataFrame.
    :param attribute_key: Attribute to filter.
    :param values: Admitted or forbidden values.
    :param level: Specifies how the filter should be applied ('case' filters the cases where at least one occurrence happens; 'event' filters the events, potentially trimming the cases).
    :param retain: Specifies if the values should be kept or removed.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_event_attribute_values(
            dataframe,
            'concept:name',
            ['Act. A', 'Act. Z'],
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, case_id_key=case_id_key)
    parameters[constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY] = attribute_key
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.attributes import attributes_filter
        if level == "event":
            parameters[attributes_filter.Parameters.POSITIVE] = retain
            return attributes_filter.apply_events(log, values, parameters=parameters)
        elif level == "case":
            parameters[attributes_filter.Parameters.POSITIVE] = retain
            return attributes_filter.apply(log, values, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.attributes import attributes_filter
        if level == "event":
            parameters[attributes_filter.Parameters.POSITIVE] = retain
            return attributes_filter.apply_events(log, values, parameters=parameters)
        elif level == "case":
            parameters[attributes_filter.Parameters.POSITIVE] = retain
            return attributes_filter.apply(log, values, parameters=parameters)


def filter_trace_attribute_values(
    log: Union[EventLog, pd.DataFrame],
    attribute_key: str,
    values: Union[Set[str], List[str]],
    retain: bool = True,
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters a log based on the values of a specified trace attribute.

    :param log: Event log or Pandas DataFrame.
    :param attribute_key: Attribute to filter.
    :param values: Collection of values to filter.
    :param retain: Boolean value indicating whether to keep or discard matching traces.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_trace_attribute_values(
            dataframe,
            'case:creator',
            ['Mike'],
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, case_id_key=case_id_key)
    parameters[constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY] = attribute_key
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.attributes import attributes_filter
        parameters[attributes_filter.Parameters.POSITIVE] = retain
        return attributes_filter.apply(log, values, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.attributes import attributes_filter
        parameters[attributes_filter.Parameters.POSITIVE] = retain
        return attributes_filter.apply_trace_attribute(log, values, parameters=parameters)


def filter_variants(
    log: Union[EventLog, pd.DataFrame],
    variants: Union[Set[str], List[str], List[Tuple[str]]],
    retain: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters a log based on a specified set of variants.

    :param log: Event log or Pandas DataFrame.
    :param variants: Collection of variants to filter. A variant should be specified as a list of tuples of activity names, e.g., [('a', 'b', 'c')].
    :param retain: Boolean indicating whether to retain (if True) or remove (if False) traces conforming to the specified variants.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_variants(
            dataframe,
            [('Act. A', 'Act. B', 'Act. Z'), ('Act. A', 'Act. C', 'Act. Z')],
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    from pm4py.util import variants_util
    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.variants import variants_filter
        parameters[variants_filter.Parameters.POSITIVE] = retain
        return variants_filter.apply(log, variants, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.variants import variants_filter
        parameters[variants_filter.Parameters.POSITIVE] = retain
        return variants_filter.apply(log, variants, parameters=parameters)


def filter_directly_follows_relation(
    log: Union[EventLog, pd.DataFrame],
    relations: List[str],
    retain: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Retains traces that contain any of the specified 'directly follows' relations.
    For example, if relations == [('a','b'),('a','c')] and log [<a,b,c>,<a,c,b>,<a,d,b>],
    the resulting log will contain traces describing [<a,b,c>,<a,c,b>].

    :param log: Event log or Pandas DataFrame.
    :param relations: List of activity name pairs, representing allowed or forbidden paths.
    :param retain: Boolean indicating whether the paths should be kept (if True) or removed (if False).
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_directly_follows_relation(
            dataframe,
            [('A', 'B'), ('A', 'C')],
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        from pm4py.algo.filtering.pandas.paths import paths_filter
        parameters[paths_filter.Parameters.POSITIVE] = retain
        return paths_filter.apply(log, relations, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.paths import paths_filter
        parameters[paths_filter.Parameters.POSITIVE] = retain
        return paths_filter.apply(log, relations, parameters=parameters)


def filter_eventually_follows_relation(
    log: Union[EventLog, pd.DataFrame],
    relations: List[str],
    retain: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Retains traces that contain any of the specified 'eventually follows' relations.
    For example, if relations == [('a','b'),('a','c')] and log [<a,b,c>,<a,c,b>,<a,d,b>],
    the resulting log will contain traces describing [<a,b,c>,<a,c,b>,<a,d,b>].

    :param log: Event log or Pandas DataFrame.
    :param relations: List of activity name pairs, representing allowed or forbidden paths.
    :param retain: Boolean indicating whether the paths should be kept (if True) or removed (if False).
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_eventually_follows_relation(
            dataframe,
            [('A', 'B'), ('A', 'C')],
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        from pm4py.algo.filtering.pandas.ltl import ltl_checker
        parameters[ltl_checker.Parameters.POSITIVE] = retain
        if retain:
            cases = set()
        else:
            cases = set(log[case_id_key].to_numpy().tolist())
        for path in relations:
            filt_log = ltl_checker.eventually_follows(log, path, parameters=parameters)
            this_traces = set(filt_log[case_id_key].to_numpy().tolist())
            if retain:
                cases = cases.union(this_traces)
            else:
                cases = cases.intersection(this_traces)
        return log[log[case_id_key].isin(cases)]
    else:
        from pm4py.algo.filtering.log.ltl import ltl_checker
        parameters[ltl_checker.Parameters.POSITIVE] = retain
        if retain:
            cases = set()
        else:
            cases = set(id(trace) for trace in log)
        for path in relations:
            filt_log = ltl_checker.eventually_follows(log, path, parameters=parameters)
            this_traces = set(id(trace) for trace in filt_log)
            if retain:
                cases = cases.union(this_traces)
            else:
                cases = cases.intersection(this_traces)
        filtered_log = EventLog(
            attributes=log.attributes,
            extensions=log.extensions,
            omni_present=log.omni_present,
            classifiers=log.classifiers,
            properties=log.properties
        )
        for trace in log:
            if id(trace) in cases:
                filtered_log.append(trace)
        return filtered_log


def filter_time_range(
    log: Union[EventLog, pd.DataFrame],
    dt1: str,
    dt2: str,
    mode: str = "events",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters a log based on a time interval.

    :param log: Event log or Pandas DataFrame.
    :param dt1: Left extreme of the interval.
    :param dt2: Right extreme of the interval.
    :param mode: Modality of filtering ('events', 'traces_contained', 'traces_intersecting').
                 - 'events': Any event that fits the time frame is retained.
                 - 'traces_contained': Any trace completely contained in the timeframe is retained.
                 - 'traces_intersecting': Any trace intersecting with the timeframe is retained.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe1 = pm4py.filter_time_range(
            dataframe,
            '2010-01-01 00:00:00',
            '2011-01-01 00:00:00',
            mode='traces_contained',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
        filtered_dataframe2 = pm4py.filter_time_range(
            dataframe,
            '2010-01-01 00:00:00',
            '2011-01-01 00:00:00',
            mode='traces_intersecting',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
        filtered_dataframe3 = pm4py.filter_time_range(
            dataframe,
            '2010-01-01 00:00:00',
            '2011-01-01 00:00:00',
            mode='events',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(log, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        from pm4py.algo.filtering.pandas.timestamp import timestamp_filter
        if mode == "events":
            return timestamp_filter.apply_events(log, dt1, dt2, parameters=properties)
        elif mode == "traces_contained":
            return timestamp_filter.filter_traces_contained(log, dt1, dt2, parameters=properties)
        elif mode == "traces_intersecting":
            return timestamp_filter.filter_traces_intersecting(log, dt1, dt2, parameters=properties)
        else:
            if constants.SHOW_INTERNAL_WARNINGS:
                warnings.warn(f"Mode provided: {mode} is not recognized; original log returned!")
            return log
    else:
        from pm4py.algo.filtering.log.timestamp import timestamp_filter
        if mode == "events":
            return timestamp_filter.apply_events(log, dt1, dt2, parameters=properties)
        elif mode == "traces_contained":
            return timestamp_filter.filter_traces_contained(log, dt1, dt2, parameters=properties)
        elif mode == "traces_intersecting":
            return timestamp_filter.filter_traces_intersecting(log, dt1, dt2, parameters=properties)
        else:
            if constants.SHOW_INTERNAL_WARNINGS:
                warnings.warn(f"Mode provided: {mode} is not recognized; original log returned!")
            return log


def filter_between(
    log: Union[EventLog, pd.DataFrame],
    act1: Union[str, List[str]],
    act2: Union[str, List[str]],
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Finds all the sub-cases leading from an event with activity "act1" to an event with activity "act2" in the log,
    and returns a log containing only them.

    Example:

    Log
    A B C D E F
    A B E F C
    A B F C B C B E F C

    act1 = B
    act2 = C

    Returned sub-cases:
    B C (from the first case)
    B E F C (from the second case)
    B F C (from the third case)
    B C (from the third case)
    B E F C (from the third case)

    :param log: Event log or Pandas DataFrame.
    :param act1: Source activity or collection of activities.
    :param act2: Target activity or collection of activities.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_between(
            dataframe,
            'A',
            'D',
            activity_key='concept:name',
            case_id_key='case:concept:name',
            timestamp_key='time:timestamp'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.between import between_filter
        return between_filter.apply(log, act1, act2, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.between import between_filter
        return between_filter.apply(log, act1, act2, parameters=parameters)


def filter_case_size(
    log: Union[EventLog, pd.DataFrame],
    min_size: int,
    max_size: int,
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the event log, keeping cases that have a length (number of events) between min_size and max_size.

    :param log: Event log or Pandas DataFrame.
    :param min_size: Minimum allowed number of events.
    :param max_size: Maximum allowed number of events.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_case_size(
            dataframe,
            5,
            10,
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.cases import case_filter
        case_id = parameters[constants.PARAMETER_CONSTANT_CASEID_KEY] if constants.PARAMETER_CONSTANT_CASEID_KEY in parameters else constants.CASE_CONCEPT_NAME
        return case_filter.filter_on_case_size(log, case_id, min_size, max_size)
    else:
        from pm4py.algo.filtering.log.cases import case_filter
        return case_filter.filter_on_case_size(log, min_size, max_size)


def filter_case_performance(
    log: Union[EventLog, pd.DataFrame],
    min_performance: float,
    max_performance: float,
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the event log, keeping cases that have a duration (the timestamp of the last event minus the timestamp
    of the first event) between min_performance and max_performance.

    :param log: Event log or Pandas DataFrame.
    :param min_performance: Minimum allowed case duration.
    :param max_performance: Maximum allowed case duration.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_case_performance(
            dataframe,
            3600.0,
            86400.0,
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.cases import case_filter
        return case_filter.filter_case_performance(log, min_performance, max_performance, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.cases import case_filter
        return case_filter.filter_case_performance(log, min_performance, max_performance, parameters=parameters)


def filter_activities_rework(
    log: Union[EventLog, pd.DataFrame],
    activity: str,
    min_occurrences: int = 2,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the event log, keeping cases where the specified activity occurs at least min_occurrences times.

    :param log: Event log or Pandas DataFrame.
    :param activity: Activity to consider.
    :param min_occurrences: Minimum desired number of occurrences.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_activities_rework(
            dataframe,
            'Approve Order',
            2,
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    parameters["min_occurrences"] = min_occurrences
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.rework import rework_filter
        return rework_filter.apply(log, activity, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.rework import rework_filter
        return rework_filter.apply(log, activity, parameters=parameters)


def filter_paths_performance(
    log: Union[EventLog, pd.DataFrame],
    path: Tuple[str, str],
    min_performance: float,
    max_performance: float,
    keep: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the event log based on the performance of specified paths.

    - If keep=True, retains cases having the specified path (tuple of 2 activities) with a duration between min_performance and max_performance.
    - If keep=False, discards cases having the specified path with a duration between min_performance and max_performance.

    :param log: Event log or Pandas DataFrame.
    :param path: Tuple of two activities (source_activity, target_activity).
    :param min_performance: Minimum allowed performance of the path.
    :param max_performance: Maximum allowed performance of the path.
    :param keep: Boolean indicating whether to keep (if True) or discard (if False) the cases with the specified performance.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_paths_performance(
            dataframe,
            ('A', 'D'),
            3600.0,
            86400.0,
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    parameters["positive"] = keep
    path = tuple(path)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.paths import paths_filter
        return paths_filter.apply_performance(log, path, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.paths import paths_filter
        return paths_filter.apply_performance(log, path, parameters=parameters)


def filter_variants_top_k(
    log: Union[EventLog, pd.DataFrame],
    k: int,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Keeps the top-k variants of the log.

    :param log: Event log or Pandas DataFrame.
    :param k: Number of variants to keep.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_variants_top_k(
            dataframe,
            5,
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.variants import variants_filter
        return variants_filter.filter_variants_top_k(log, k, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.variants import variants_filter
        return variants_filter.filter_variants_top_k(log, k, parameters=parameters)


def filter_variants_by_coverage_percentage(
    log: Union[EventLog, pd.DataFrame],
    min_coverage_percentage: float,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the variants of the log based on a coverage percentage.
    For example, if min_coverage_percentage=0.4 and the log has 1000 cases with:
    - 500 cases of variant 1,
    - 400 cases of variant 2,
    - 100 cases of variant 3,
    the filter keeps only the traces of variant 1 and variant 2.

    :param log: Event log or Pandas DataFrame.
    :param min_coverage_percentage: Minimum allowed percentage of coverage.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_variants_by_coverage_percentage(
            dataframe,
            0.1,
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.variants import variants_filter
        return variants_filter.filter_variants_by_coverage_percentage(log, min_coverage_percentage, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.variants import variants_filter
        return variants_filter.filter_variants_by_coverage_percentage(log, min_coverage_percentage, parameters=parameters)


def filter_prefixes(
    log: Union[EventLog, pd.DataFrame],
    activity: str,
    strict: bool = True,
    first_or_last: str = "first",
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the log, keeping the prefixes leading up to a given activity.
    For example, for a log with traces:
    - A,B,C,D
    - A,B,Z,A,B,C,D
    - A,B,C,D,C,E,C,F

    The prefixes to "C" are respectively:
    - A,B
    - A,B,Z,A,B
    - A,B

    :param log: Event log or Pandas DataFrame.
    :param activity: Target activity for the filter.
    :param strict: Applies the filter strictly, cutting the occurrences of the selected activity.
    :param first_or_last: Decides if the first or last occurrence of an activity should be selected as the baseline for the filter.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_prefixes(
            dataframe,
            'Act. C',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    parameters["strict"] = strict
    parameters["first_or_last"] = first_or_last

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.prefixes import prefix_filter
        return prefix_filter.apply(log, activity, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.prefixes import prefix_filter
        return prefix_filter.apply(log, activity, parameters=parameters)


def filter_suffixes(
    log: Union[EventLog, pd.DataFrame],
    activity: str,
    strict: bool = True,
    first_or_last: str = "first",
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the log, keeping the suffixes starting from a given activity.
    For example, for a log with traces:
    - A,B,C,D
    - A,B,Z,A,B,C,D
    - A,B,C,D,C,E,C,F

    The suffixes from "C" are respectively:
    - D
    - D
    - D,C,E,C,F

    :param log: Event log or Pandas DataFrame.
    :param activity: Target activity for the filter.
    :param strict: Applies the filter strictly, cutting the occurrences of the selected activity.
    :param first_or_last: Decides if the first or last occurrence of an activity should be selected as the baseline for the filter.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_suffixes(
            dataframe,
            'Act. C',
            activity_key='concept:name',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    parameters["strict"] = strict
    parameters["first_or_last"] = first_or_last

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.suffixes import suffix_filter
        return suffix_filter.apply(log, activity, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.suffixes import suffix_filter
        return suffix_filter.apply(log, activity, parameters=parameters)


def filter_ocel_event_attribute(
    ocel: OCEL,
    attribute_key: str,
    attribute_values: Collection[Any],
    positive: bool = True
) -> OCEL:
    """
    Filters the object-centric event log based on the provided event attribute values.

    :param ocel: Object-centric event log.
    :param attribute_key: Attribute at the event level to filter.
    :param attribute_values: Collection of attribute values to keep or remove.
    :param positive: Determines whether the values should be kept (True) or removed (False).
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_event_attribute(
            ocel,
            'ocel:activity',
            ['A', 'B', 'D']
        )
    """
    from pm4py.algo.filtering.ocel import event_attributes

    return event_attributes.apply(
        ocel,
        attribute_values,
        parameters={
            event_attributes.Parameters.ATTRIBUTE_KEY: attribute_key,
            event_attributes.Parameters.POSITIVE: positive
        }
    )


def filter_ocel_object_attribute(
    ocel: OCEL,
    attribute_key: str,
    attribute_values: Collection[Any],
    positive: bool = True
) -> OCEL:
    """
    Filters the object-centric event log based on the provided object attribute values.

    :param ocel: Object-centric event log.
    :param attribute_key: Attribute at the object level to filter.
    :param attribute_values: Collection of attribute values to keep or remove.
    :param positive: Determines whether the values should be kept (True) or removed (False).
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_object_attribute(
            ocel,
            'ocel:type',
            ['order']
        )
    """
    from pm4py.algo.filtering.ocel import object_attributes

    return object_attributes.apply(
        ocel,
        attribute_values,
        parameters={
            object_attributes.Parameters.ATTRIBUTE_KEY: attribute_key,
            object_attributes.Parameters.POSITIVE: positive
        }
    )


def filter_ocel_object_types_allowed_activities(
    ocel: OCEL,
    correspondence_dict: Dict[str, Collection[str]]
) -> OCEL:
    """
    Filters an object-centric event log, keeping only the specified object types with the specified set of allowed activities.

    :param ocel: Object-centric event log.
    :param correspondence_dict: Dictionary containing, for every object type of interest, a collection of allowed activities.
                                Example: {"order": ["Create Order"], "element": ["Create Order", "Create Delivery"]}.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_object_types_allowed_activities(
            ocel,
            {'order': ['create order', 'pay order'], 'item': ['create item', 'deliver item']}
        )
    """
    from pm4py.algo.filtering.ocel import activity_type_matching

    return activity_type_matching.apply(ocel, correspondence_dict)


def filter_ocel_object_per_type_count(
    ocel: OCEL,
    min_num_obj_type: Dict[str, int]
) -> OCEL:
    """
    Filters the events of the object-centric logs that are related to at least the specified number of objects per type.

    Example:
    pm4py.filter_object_per_type_count(ocel, {"order": 1, "element": 2})

    Would keep the following events:

      ocel:eid ocel:timestamp ocel:activity ocel:type:element ocel:type:order
    0       e1     1980-01-01  Create Order  [i4, i1, i3, i2]            [o1]
    1      e11     1981-01-01  Create Order          [i6, i5]            [o2]
    2      e14     1981-01-04  Create Order          [i8, i7]            [o3]

    :param ocel: Object-centric event log.
    :param min_num_obj_type: Minimum number of objects per type.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_object_per_type_count(
            ocel,
            {'order': 1, 'element': 2}
        )
    """
    from pm4py.algo.filtering.ocel import objects_ot_count

    return objects_ot_count.apply(ocel, min_num_obj_type)


def filter_ocel_start_events_per_object_type(
    ocel: OCEL,
    object_type: str
) -> OCEL:
    """
    Filters the events in which a new object of the given object type is spawned.
    For example, an event with activity "Create Order" might spawn new orders.

    :param ocel: Object-centric event log.
    :param object_type: Object type to consider.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_start_events_per_object_type(
            ocel,
            'delivery'
        )
    """
    from pm4py.algo.filtering.ocel import ot_endpoints
    return ot_endpoints.filter_start_events_per_object_type(ocel, object_type)


def filter_ocel_end_events_per_object_type(
    ocel: OCEL,
    object_type: str
) -> OCEL:
    """
    Filters the events in which an object of the given object type terminates its lifecycle.
    For example, an event with activity "Pay Order" might terminate an order.

    :param ocel: Object-centric event log.
    :param object_type: Object type to consider.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_end_events_per_object_type(
            ocel,
            'delivery'
        )
    """
    from pm4py.algo.filtering.ocel import ot_endpoints
    return ot_endpoints.filter_end_events_per_object_type(ocel, object_type)


def filter_ocel_events_timestamp(
    ocel: OCEL,
    min_timest: Union[datetime.datetime, str],
    max_timest: Union[datetime.datetime, str],
    timestamp_key: str = "ocel:timestamp"
) -> OCEL:
    """
    Filters the object-centric event log, keeping events within the provided timestamp range.

    :param ocel: Object-centric event log.
    :param min_timest: Left extreme of the allowed timestamp interval (format: YYYY-mm-dd HH:MM:SS).
    :param max_timest: Right extreme of the allowed timestamp interval (format: YYYY-mm-dd HH:MM:SS).
    :param timestamp_key: The attribute to use as timestamp (default: ocel:timestamp).
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_events_timestamp(
            ocel,
            '1990-01-01 00:00:00',
            '2010-01-01 00:00:00'
        )
    """
    from pm4py.algo.filtering.ocel import event_attributes
    return event_attributes.apply_timestamp(
        ocel,
        min_timest,
        max_timest,
        parameters={"pm4py:param:timestamp_key": timestamp_key}
    )


def filter_four_eyes_principle(
    log: Union[EventLog, pd.DataFrame],
    activity1: str,
    activity2: str,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    resource_key: str = "org:resource",
    keep_violations: bool = False
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters out the cases of the log that violate the four-eyes principle on the provided activities.

    :param log: Event log or Pandas DataFrame.
    :param activity1: First activity.
    :param activity2: Second activity.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :param resource_key: Attribute to be used as resource.
    :param keep_violations: Boolean indicating whether to discard (if False) or retain (if True) the violations.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_four_eyes_principle(
            dataframe,
            'Act. A',
            'Act. B',
            activity_key='concept:name',
            resource_key='org:resource',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key, resource_key=resource_key)
    properties["positive"] = not keep_violations

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)

        from pm4py.algo.filtering.pandas.ltl import ltl_checker
        return ltl_checker.four_eyes_principle(log, activity1, activity2, parameters=properties)
    else:
        from pm4py.algo.filtering.log.ltl import ltl_checker
        return ltl_checker.four_eyes_principle(log, activity1, activity2, parameters=properties)


def filter_activity_done_different_resources(
    log: Union[EventLog, pd.DataFrame],
    activity: str,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name",
    resource_key: str = "org:resource",
    keep_violations: bool = True
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters the cases where an activity is performed by different resources multiple times.

    :param log: Event log or Pandas DataFrame.
    :param activity: Activity to consider.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :param resource_key: Attribute to be used as resource.
    :param keep_violations: Boolean indicating whether to discard (if False) or retain (if True) the violations.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        filtered_dataframe = pm4py.filter_activity_done_different_resources(
            dataframe,
            'Act. A',
            activity_key='concept:name',
            resource_key='org:resource',
            timestamp_key='time:timestamp',
            case_id_key='case:concept:name'
        )
    """
    __event_log_deprecation_warning(log)

    properties = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key, resource_key=resource_key)
    properties["positive"] = keep_violations

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)

        from pm4py.algo.filtering.pandas.ltl import ltl_checker
        return ltl_checker.attr_value_different_persons(log, activity, parameters=properties)
    else:
        from pm4py.algo.filtering.log.ltl import ltl_checker
        return ltl_checker.attr_value_different_persons(log, activity, parameters=properties)


def filter_trace_segments(
    log: Union[EventLog, pd.DataFrame],
    admitted_traces: List[List[str]],
    positive: bool = True,
    activity_key: str = "concept:name",
    timestamp_key: str = "time:timestamp",
    case_id_key: str = "case:concept:name"
) -> Union[EventLog, pd.DataFrame]:
    """
    Filters an event log based on a set of trace segments. A trace is a sequence of activities and "..."
    where:
    - "..." before an activity indicates that other activities can precede the given activity.
    - "..." after an activity indicates that other activities can follow the given activity.

    Examples:
    - pm4py.filter_trace_segments(log, [["A", "B"]]) retains only cases with the exact process variant A,B.
    - pm4py.filter_trace_segments(log, [["...", "A", "B"]]) retains only cases ending with activities A,B.
    - pm4py.filter_trace_segments(log, [["A", "B", "..."]]) retains only cases starting with activities A,B.
    - pm4py.filter_trace_segments(log, [["...", "A", "B", "C", "..."], ["...", "D", "E", "F", "..."]]) retains cases where:
        - At any point, there is A followed by B followed by C,
        - And at any other point, there is D followed by E followed by F.

    :param log: Event log or Pandas DataFrame.
    :param admitted_traces: Collection of trace segments to admit based on the criteria above.
    :param positive: Boolean indicating whether to keep (if True) or discard (if False) the cases satisfying the filter.
    :param activity_key: Attribute to be used for the activity.
    :param timestamp_key: Attribute to be used for the timestamp.
    :param case_id_key: Attribute to be used as case identifier.
    :return: Filtered event log or Pandas DataFrame.

    .. code-block:: python3

        import pm4py

        log = pm4py.read_xes("tests/input_data/running-example.xes")

        filtered_log = pm4py.filter_trace_segments(
            log,
            [["...", "check ticket", "decide", "reinitiate request", "..."]]
        )
        print(filtered_log)
    """
    __event_log_deprecation_warning(log)

    parameters = get_properties(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
    parameters["positive"] = positive

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_key=case_id_key)
        from pm4py.algo.filtering.pandas.traces import trace_filter
        return trace_filter.apply(log, admitted_traces, parameters=parameters)
    else:
        from pm4py.algo.filtering.log.traces import trace_filter
        return trace_filter.apply(log, admitted_traces, parameters=parameters)


def filter_ocel_object_types(
    ocel: OCEL,
    obj_types: Collection[str],
    positive: bool = True,
    level: int = 1
) -> OCEL:
    """
    Filters the object types of an object-centric event log.

    :param ocel: Object-centric event log.
    :param obj_types: Object types to keep or remove.
    :param positive: Boolean indicating whether to keep (True) or remove (False) the specified object types.
    :param level: Recursively expands the set of object identifiers until the specified level.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('log.jsonocel')
        filtered_ocel = pm4py.filter_ocel_object_types(
            ocel,
            ['order']
        )
    """
    from copy import copy
    from pm4py.objects.ocel.util import filtering_utils
    if level == 1:
        filtered_ocel = copy(ocel)
        if positive:
            filtered_ocel.objects = filtered_ocel.objects[filtered_ocel.objects[filtered_ocel.object_type_column].isin(obj_types)]
        else:
            filtered_ocel.objects = filtered_ocel.objects[~filtered_ocel.objects[filtered_ocel.object_type_column].isin(obj_types)]
        return filtering_utils.propagate_object_filtering(filtered_ocel)
    else:
        object_ids = pandas_utils.format_unique(
            ocel.objects[ocel.objects[ocel.object_type_column].isin(obj_types)][ocel.object_id_column].unique()
        )
        return filter_ocel_objects(ocel, object_ids, level=level, positive=positive)


def filter_ocel_objects(
    ocel: OCEL,
    object_identifiers: Collection[str],
    positive: bool = True,
    level: int = 1
) -> OCEL:
    """
    Filters the object identifiers of an object-centric event log.

    :param ocel: Object-centric event log.
    :param object_identifiers: Object identifiers to keep or remove.
    :param positive: Boolean indicating whether to keep (True) or remove (False) the specified object identifiers.
    :param level: Recursively expands the set of object identifiers until the specified level.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('log.jsonocel')
        filtered_ocel = pm4py.filter_ocel_objects(
            ocel,
            ['o1'],
            level=1
        )
    """
    object_identifiers = set(object_identifiers)
    if level > 1:
        ev_rel_obj = ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column].agg(list).to_dict()
        objects_ids = ocel.objects[ocel.object_id_column].to_numpy().tolist()
        graph = {o: set() for o in objects_ids}
        for ev in ev_rel_obj:
            rel_obj = ev_rel_obj[ev]
            for o1 in rel_obj:
                for o2 in rel_obj:
                    if o1 != o2:
                        graph[o1].add(o2)
        while level > 1:
            curr = list(object_identifiers)
            for el in curr:
                for el2 in graph[el]:
                    object_identifiers.add(el2)
            level = level - 1
    from copy import copy
    from pm4py.objects.ocel.util import filtering_utils
    filtered_ocel = copy(ocel)
    if positive:
        filtered_ocel.objects = filtered_ocel.objects[filtered_ocel.objects[ocel.object_id_column].isin(object_identifiers)]
    else:
        filtered_ocel.objects = filtered_ocel.objects[~filtered_ocel.objects[ocel.object_id_column].isin(object_identifiers)]
    return filtering_utils.propagate_object_filtering(filtered_ocel)


def filter_ocel_events(
    ocel: OCEL,
    event_identifiers: Collection[str],
    positive: bool = True
) -> OCEL:
    """
    Filters the event identifiers of an object-centric event log.

    :param ocel: Object-centric event log.
    :param event_identifiers: Event identifiers to keep or remove.
    :param positive: Boolean indicating whether to keep (True) or remove (False) the specified event identifiers.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('log.jsonocel')
        filtered_ocel = pm4py.filter_ocel_events(
            ocel,
            ['e1']
        )
    """
    from copy import copy
    from pm4py.objects.ocel.util import filtering_utils
    filtered_ocel = copy(ocel)
    if positive:
        filtered_ocel.events = filtered_ocel.events[filtered_ocel.events[ocel.event_id_column].isin(event_identifiers)]
    else:
        filtered_ocel.events = filtered_ocel.events[~filtered_ocel.events[ocel.event_id_column].isin(event_identifiers)]
    return filtering_utils.propagate_event_filtering(filtered_ocel)


def filter_ocel_activities_connected_object_type(ocel: OCEL, object_type: str) -> OCEL:
    """
    Filter an OCEL on the set of activities executed on objects of the given object type.

    :param ocel: object-centric event log
    :param object_type: object type
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel2("tests/input_data/ocel/ocel20_example.xmlocel")
        filtered_ocel = pm4py.filter_ocel_activities_connected_object_type(ocel, "Purchase Order")
        print(filtered_ocel)
    """
    from copy import copy
    from pm4py.objects.ocel.util import filtering_utils
    relations = ocel.relations[ocel.relations[ocel.object_type_column] == object_type]
    activities = relations[ocel.event_activity].unique()
    filtered_ocel = copy(ocel)
    filtered_ocel.relations = filtered_ocel.relations[filtered_ocel.relations[ocel.event_activity].isin(activities)]

    return filtering_utils.propagate_relations_filtering(filtered_ocel)


def filter_ocel_cc_object(
    ocel: OCEL,
    object_id: str,
    conn_comp: Optional[List[List[str]]] = None,
    return_conn_comp: bool = False
) -> Union[OCEL, Tuple[OCEL, List[List[str]]]]:
    """
    Returns the connected component of the object-centric event log to which the specified object belongs.

    :param ocel: Object-centric event log.
    :param object_id: Object identifier.
    :param conn_comp: (Optional) Precomputed connected components of the OCEL objects.
    :param return_conn_comp: If True, returns the filtered OCEL along with the computed connected components.
    :return: Filtered OCEL, optionally with the list of connected components.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('log.jsonocel')
        filtered_ocel = pm4py.filter_ocel_cc_object(
            ocel,
            'order1'
        )
    """
    if conn_comp is None:
        from pm4py.algo.transformation.ocel.graphs import object_interaction_graph

        g0 = object_interaction_graph.apply(ocel)
        g = nx_utils.Graph()

        for edge in g0:
            g.add_edge(edge[0], edge[1])

        conn_comp = list(nx_utils.connected_components(g))

    for cc in conn_comp:
        if object_id in cc:
            if return_conn_comp:
                return filter_ocel_objects(ocel, cc), conn_comp
            else:
                return filter_ocel_objects(ocel, cc)

    if return_conn_comp:
        return filter_ocel_objects(ocel, [object_id]), conn_comp
    else:
        return filter_ocel_objects(ocel, [object_id])


def filter_ocel_cc_length(
    ocel: OCEL,
    min_cc_length: int,
    max_cc_length: int
) -> OCEL:
    """
    Keeps only the objects in an OCEL belonging to a connected component with a length
    falling within the specified range.

    Reference:
    Adams, Jan Niklas, et al. "Defining cases and variants for object-centric event data." 
    2022 4th International Conference on Process Mining (ICPM). IEEE, 2022.

    :param ocel: Object-centric event log.
    :param min_cc_length: Minimum allowed length for the connected component.
    :param max_cc_length: Maximum allowed length for the connected component.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        filtered_ocel = pm4py.filter_ocel_cc_length(
            ocel,
            2,
            10
        )
    """
    from pm4py.algo.transformation.ocel.graphs import object_interaction_graph

    g0 = object_interaction_graph.apply(ocel)
    g = nx_utils.Graph()

    for edge in g0:
        g.add_edge(edge[0], edge[1])

    conn_comp = list(nx_utils.connected_components(g))
    conn_comp = [x for x in conn_comp if min_cc_length <= len(x) <= max_cc_length]
    objs = [y for x in conn_comp for y in x]

    return filter_ocel_objects(ocel, objs)


def filter_ocel_cc_otype(
    ocel: OCEL,
    otype: str,
    positive: bool = True
) -> OCEL:
    """
    Filters the objects belonging to connected components that have at least one object of the specified type.

    Reference:
    Adams, Jan Niklas, et al. "Defining cases and variants for object-centric event data." 
    2022 4th International Conference on Process Mining (ICPM). IEEE, 2022.

    :param ocel: Object-centric event log.
    :param otype: Object type to consider.
    :param positive: Boolean indicating whether to keep (True) or discard (False) the objects in these components.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('log.jsonocel')
        filtered_ocel = pm4py.filter_ocel_cc_otype(
            ocel,
            'order'
        )
    """
    if positive:
        objs = set(
            ocel.objects[ocel.objects[ocel.object_type_column] == otype][ocel.object_id_column]
        )
    else:
        objs = set(
            ocel.objects[~(ocel.objects[ocel.object_type_column] == otype)][ocel.object_id_column]
        )

    from pm4py.algo.transformation.ocel.graphs import object_interaction_graph

    g0 = object_interaction_graph.apply(ocel)
    g = nx_utils.Graph()

    for edge in g0:
        g.add_edge(edge[0], edge[1])

    conn_comp = list(nx_utils.connected_components(g))
    conn_comp = [x for x in conn_comp if len(set(x).intersection(objs)) > 0]

    objs = [y for x in conn_comp for y in x]

    return filter_ocel_objects(ocel, objs)


def filter_ocel_cc_activity(
    ocel: OCEL,
    activity: str
) -> OCEL:
    """
    Filters the objects belonging to connected components that include at least one event with the specified activity.

    Reference:
    Adams, Jan Niklas, et al. "Defining cases and variants for object-centric event data." 
    2022 4th International Conference on Process Mining (ICPM). IEEE, 2022.

    :param ocel: Object-centric event log.
    :param activity: Activity to consider.
    :return: Filtered OCEL.

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('log.jsonocel')
        filtered_ocel = pm4py.filter_ocel_cc_activity(
            ocel,
            'Create Order'
        )
    """
    evs = ocel.events[ocel.events[ocel.event_activity] == activity][ocel.event_id_column].to_numpy().tolist()
    objs = pandas_utils.format_unique(
        ocel.relations[ocel.relations[ocel.event_id_column].isin(evs)][ocel.object_id_column].unique()
    )

    from pm4py.algo.transformation.ocel.graphs import object_interaction_graph

    g0 = object_interaction_graph.apply(ocel)
    g = nx_utils.Graph()

    for edge in g0:
        g.add_edge(edge[0], edge[1])

    conn_comp = list(nx_utils.connected_components(g))
    conn_comp = [x for x in conn_comp if len(set(x).intersection(objs)) > 0]

    objs = [y for x in conn_comp for y in x]

    return filter_ocel_objects(ocel, objs)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional
import pandas as pd
from pm4py.objects.ocel.obj import OCEL


def extract_log_outlook_mails() -> pd.DataFrame:
    """
    Extracts the history of conversations from the local instance of Microsoft Outlook
    running on the current computer.

    Columns:
    - **CASE ID (case:concept:name)**: Identifier of the conversation.
    - **ACTIVITY (concept:name)**: Activity performed in the current item (e.g., send e-mail, receive e-mail, refuse meeting).
    - **TIMESTAMP (time:timestamp)**: Timestamp of creation of the item in Outlook.
    - **RESOURCE (org:resource)**: Sender of the current item.

    See also:
    * [MailItem Properties](https://learn.microsoft.com/en-us/dotnet/api/microsoft.office.interop.outlook.mailitem?redirectedfrom=MSDN&view=outlook-pia#properties_)
    * [OlObjectClass Enumeration](https://learn.microsoft.com/en-us/dotnet/api/microsoft.office.interop.outlook.olobjectclass?view=outlook-pia)

    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_outlook_mails()
    """
    from pm4py.algo.connectors.variants import outlook_mail_extractor
    return outlook_mail_extractor.apply()


def extract_log_outlook_calendar(email_user: Optional[str] = None, calendar_id: int = 9) -> pd.DataFrame:
    """
    Extracts the history of calendar events (creation, update, start, end)
    into a Pandas DataFrame from the local Outlook instance running on the current computer.

    Columns:
    - **CASE ID (case:concept:name)**: Identifier of the meeting.
    - **ACTIVITY (concept:name)**: One of the following activities: Meeting Created, Last Change of Meeting, Meeting Started, Meeting Completed.
    - **TIMESTAMP (time:timestamp)**: Timestamp of the event.
    - **case:subject**: Subject of the meeting.

    :param email_user: (optional) E-mail address from which the (shared) calendar should be extracted.
    :param calendar_id: Identifier of the calendar for the given user (default: 9).

    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        # Extract using default parameters
        dataframe = pm4py.connectors.extract_log_outlook_calendar()

        # Extract using a specific email user
        dataframe = pm4py.connectors.extract_log_outlook_calendar("vacation-calendar@workplace.eu")
    """
    from pm4py.algo.connectors.variants import outlook_calendar
    parameters = {}
    parameters[outlook_calendar.Parameters.EMAIL_USER] = email_user
    parameters[outlook_calendar.Parameters.CALENDAR_ID] = calendar_id
    return outlook_calendar.apply(parameters=parameters)


def extract_log_windows_events() -> pd.DataFrame:
    """
    Extracts a process mining DataFrame from all events recorded in the Windows registry.

    Columns:
    - **CASE ID (case:concept:name)**: Name of the computer emitting the events.
    - **ACTIVITY (concept:name)**: Concatenation of the source name of the event and the event identifier.
      (See [Win32_NTLogEvent](https://learn.microsoft.com/en-us/previous-versions/windows/desktop/eventlogprov/win32-ntlogevent))
    - **TIMESTAMP (time:timestamp)**: Timestamp of event generation.
    - **RESOURCE (org:resource)**: Username involved in the event.

    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_windows_events()
    """
    from pm4py.algo.connectors.variants import windows_events
    return windows_events.apply()


def extract_log_chrome_history(history_db_path: Optional[str] = None) -> pd.DataFrame:
    """
    Extracts a DataFrame containing the navigation history of Google Chrome.
    Please ensure that Google Chrome history is closed when extracting.

    Columns:
    - **CASE ID (case:concept:name)**: Identifier of the extracted profile.
    - **ACTIVITY (concept:name)**: Complete path of the website, excluding GET arguments.
    - **TIMESTAMP (time:timestamp)**: Timestamp of the visit.

    :param history_db_path: Path to the Google Chrome history database (default: location of the Windows folder).
    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_chrome_history()
    """
    from pm4py.algo.connectors.variants import chrome_history
    parameters = {}
    if history_db_path is not None:
        parameters[chrome_history.Parameters.HISTORY_DB_PATH] = history_db_path
    return chrome_history.apply(parameters=parameters)


def extract_log_firefox_history(history_db_path: Optional[str] = None) -> pd.DataFrame:
    """
    Extracts a DataFrame containing the navigation history of Mozilla Firefox.
    Please ensure that Mozilla Firefox history is closed when extracting.

    Columns:
    - **CASE ID (case:concept:name)**: Identifier of the extracted profile.
    - **ACTIVITY (concept:name)**: Complete path of the website, excluding GET arguments.
    - **TIMESTAMP (time:timestamp)**: Timestamp of the visit.

    :param history_db_path: Path to the Mozilla Firefox history database (default: location of the Windows folder).
    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_firefox_history()
    """
    from pm4py.algo.connectors.variants import firefox_history
    parameters = {}
    if history_db_path is not None:
        parameters[firefox_history.Parameters.HISTORY_DB_PATH] = history_db_path
    return firefox_history.apply(parameters=parameters)


def extract_log_github(owner: str = "pm4py", repo: str = "pm4py-core", auth_token: Optional[str] = None) -> pd.DataFrame:
    """
    Extracts a DataFrame containing the history of issues from a GitHub repository.
    Due to API rate limits for public and registered users, only a subset of events may be returned.

    :param owner: Owner of the repository (e.g., pm4py).
    :param repo: Name of the repository (e.g., pm4py-core).
    :param auth_token: Authorization token.
    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_github(owner='pm4py', repo='pm4py-core')
    """
    from pm4py.algo.connectors.variants import github_repo
    parameters = {}
    parameters[github_repo.Parameters.OWNER] = owner
    parameters[github_repo.Parameters.REPOSITORY] = repo
    parameters[github_repo.Parameters.AUTH_TOKEN] = auth_token
    return github_repo.apply(parameters)


def extract_log_camunda_workflow(connection_string: str) -> pd.DataFrame:
    """
    Extracts a DataFrame from the Camunda workflow system. In addition to traditional columns,
    the process ID of the process in Camunda is included.

    :param connection_string: ODBC connection string to the Camunda database.
    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_camunda_workflow(
            'Driver={PostgreSQL Unicode(x64)};SERVER=127.0.0.3;DATABASE=process-engine;UID=xx;PWD=yy'
        )
    """
    from pm4py.algo.connectors.variants import camunda_workflow
    parameters = {}
    parameters[camunda_workflow.Parameters.CONNECTION_STRING] = connection_string
    return camunda_workflow.apply(None, parameters=parameters)


def extract_log_sap_o2c(connection_string: str, prefix: str = "") -> pd.DataFrame:
    """
    Extracts a DataFrame for the SAP Order-to-Cash (O2C) process.

    :param connection_string: ODBC connection string to the SAP database.
    :param prefix: Prefix for the tables (e.g., SAPSR3.).
    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_sap_o2c(
            'Driver={Oracle in instantclient_21_6};DBQ=127.0.0.3:1521/ZIB;UID=xx;PWD=yy'
        )
    """
    from pm4py.algo.connectors.variants import sap_o2c
    parameters = {}
    parameters[sap_o2c.Parameters.CONNECTION_STRING] = connection_string
    parameters[sap_o2c.Parameters.PREFIX] = prefix
    return sap_o2c.apply(None, parameters=parameters)


def extract_log_sap_accounting(connection_string: str, prefix: str = "") -> pd.DataFrame:
    """
    Extracts a DataFrame for the SAP Accounting process.

    :param connection_string: ODBC connection string to the SAP database.
    :param prefix: Prefix for the tables (e.g., SAPSR3.).
    :rtype: ``pd.DataFrame``

    .. code-block:: python3
        import pm4py

        dataframe = pm4py.connectors.extract_log_sap_accounting(
            'Driver={Oracle in instantclient_21_6};DBQ=127.0.0.3:1521/ZIB;UID=xx;PWD=yy'
        )
    """
    from pm4py.algo.connectors.variants import sap_accounting
    parameters = {}
    parameters[sap_accounting.Parameters.CONNECTION_STRING] = connection_string
    parameters[sap_accounting.Parameters.PREFIX] = prefix
    return sap_accounting.apply(None, parameters=parameters)


def extract_ocel_outlook_mails() -> OCEL:
    """
    Extracts the history of conversations from the local instance of Microsoft Outlook
    running on the current computer as an object-centric event log.

    Columns:
    - **ACTIVITY (ocel:activity)**: Activity performed in the current item (e.g., send e-mail, receive e-mail, refuse meeting).
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of creation of the item in Outlook.

    Object Types:
    - **org:resource**: Sender of the mail.
    - **recipients**: List of recipients of the mail.
    - **topic**: Topic of the discussion.

    See also:
    * [MailItem Properties](https://learn.microsoft.com/en-us/dotnet/api/microsoft.office.interop.outlook.mailitem?redirectedfrom=MSDN&view=outlook-pia#properties_)
    * [OlObjectClass Enumeration](https://learn.microsoft.com/en-us/dotnet/api/microsoft.office.interop.outlook.olobjectclass?view=outlook-pia)

    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_outlook_mails()
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_outlook_mails()
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["org:resource", "recipients", "topic"]
    )


def extract_ocel_outlook_calendar(email_user: Optional[str] = None, calendar_id: int = 9) -> OCEL:
    """
    Extracts the history of calendar events (creation, update, start, end)
    as an object-centric event log from the local Outlook instance running on the current computer.

    Columns:
    - **ACTIVITY (ocel:activity)**: One of the following activities: Meeting Created, Last Change of Meeting, Meeting Started, Meeting Completed.
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of the event.

    Object Types:
    - **case:concept:name**: Identifier of the meeting.
    - **case:subject**: Subject of the meeting.

    :param email_user: (optional) E-mail address from which the (shared) calendar should be extracted.
    :param calendar_id: Identifier of the calendar for the given user (default: 9).
    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        # Extract using default parameters
        ocel = pm4py.connectors.extract_ocel_outlook_calendar()

        # Extract using a specific email user
        ocel = pm4py.connectors.extract_ocel_outlook_calendar("vacation-calendar@workplace.eu")
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_outlook_calendar(email_user, calendar_id)
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["case:concept:name", "case:subject"]
    )


def extract_ocel_windows_events() -> OCEL:
    """
    Extracts an object-centric event log from all events recorded in the Windows registry.

    Columns:
    - **ACTIVITY (ocel:activity)**: Concatenation of the source name of the event and the event identifier.
      (See [Win32_NTLogEvent](https://learn.microsoft.com/en-us/previous-versions/windows/desktop/eventlogprov/win32-ntlogevent))
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of event generation.

    Object Types:
    - **categoryString**: Translation of the subcategory. The translation is source-specific.
    - **computerName**: Name of the computer that generated the event.
    - **eventIdentifier**: Identifier of the event, specific to the source that generated the event log entry.
    - **eventType**: Event type classification (1=Error; 2=Warning; 3=Information; 4=Security Audit Success; 5=Security Audit Failure).
    - **sourceName**: Name of the source (application, service, driver, or subsystem) that generated the entry.
    - **user**: Username of the logged-on user when the event occurred. If the username cannot be determined, this will be NULL.

    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_windows_events()
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_windows_events()
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["categoryString", "computerName", "eventIdentifier", "eventType", "sourceName", "user"]
    )


def extract_ocel_chrome_history(history_db_path: Optional[str] = None) -> OCEL:
    """
    Extracts an object-centric event log containing the navigation history of Google Chrome.
    Please ensure that Google Chrome history is closed when extracting.

    Columns:
    - **ACTIVITY (ocel:activity)**: Complete path of the website, excluding GET arguments.
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of the visit.

    Object Types:
    - **case:concept:name**: Profile of Chrome used to visit the site.
    - **complete_url**: Complete URL of the website.
    - **url_wo_parameters**: Complete URL excluding the part after '?'.
    - **domain**: Domain of the visited website.

    :param history_db_path: Path to the Google Chrome history database (default: location of the Windows folder).
    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_chrome_history()
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_chrome_history(history_db_path)
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["case:concept:name", "complete_url", "url_wo_parameters", "domain"]
    )


def extract_ocel_firefox_history(history_db_path: Optional[str] = None) -> OCEL:
    """
    Extracts an object-centric event log containing the navigation history of Mozilla Firefox.
    Please ensure that Mozilla Firefox history is closed when extracting.

    Columns:
    - **ACTIVITY (ocel:activity)**: Complete path of the website, excluding GET arguments.
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of the visit.

    Object Types:
    - **case:concept:name**: Profile of Firefox used to visit the site.
    - **complete_url**: Complete URL of the website.
    - **url_wo_parameters**: Complete URL excluding the part after '?'.
    - **domain**: Domain of the visited website.

    :param history_db_path: Path to the Mozilla Firefox history database (default: location of the Windows folder).
    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_firefox_history()
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_firefox_history(history_db_path)
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["case:concept:name", "complete_url", "url_wo_parameters", "domain"]
    )


def extract_ocel_github(owner: str = "pm4py", repo: str = "pm4py-core", auth_token: Optional[str] = None) -> OCEL:
    """
    Extracts an object-centric event log containing the history of issues from a GitHub repository.
    Due to API rate limits for public and registered users, only a subset of events may be returned.

    Columns:
    - **ACTIVITY (ocel:activity)**: The event type (e.g., created, commented, closed, subscribed).
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of the event execution.

    Object Types:
    - **case:concept:name**: URL of the events related to the issue.
    - **org:resource**: Involved resource.
    - **case:repo**: Repository in which the issue was created.

    :param owner: Owner of the repository (e.g., pm4py).
    :param repo: Name of the repository (e.g., pm4py-core).
    :param auth_token: Authorization token.
    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_github(owner='pm4py', repo='pm4py-core')
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_github(owner, repo, auth_token)
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["case:concept:name", "org:resource", "case:repo"]
    )


def extract_ocel_camunda_workflow(connection_string: str) -> OCEL:
    """
    Extracts an object-centric event log from the Camunda workflow system.

    Columns:
    - **ACTIVITY (ocel:activity)**: Activity performed within Camunda.
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of the activity execution.

    Object Types:
    - **case:concept:name**: Identifier of the case.
    - **processID**: Process ID within Camunda.
    - **org:resource**: Resource involved in the activity.

    :param connection_string: ODBC connection string to the Camunda database.
    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_camunda_workflow(
            'Driver={PostgreSQL Unicode(x64)};SERVER=127.0.0.3;DATABASE=process-engine;UID=xx;PWD=yy'
        )
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_camunda_workflow(connection_string)
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["case:concept:name", "processID", "org:resource"]
    )


def extract_ocel_sap_o2c(connection_string: str, prefix: str = '') -> OCEL:
    """
    Extracts an object-centric event log for the SAP Order-to-Cash (O2C) process.

    Columns:
    - **ACTIVITY (ocel:activity)**: Activity performed in the O2C process.
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of the activity execution.

    Object Types:
    - **case:concept:name**: Identifier of the case.
    - **org:resource**: Resource involved in the activity.

    :param connection_string: ODBC connection string to the SAP database.
    :param prefix: Prefix for the tables (e.g., SAPSR3.).
    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_sap_o2c(
            'Driver={Oracle in instantclient_21_6};DBQ=127.0.0.3:1521/ZIB;UID=xx;PWD=yy'
        )
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_sap_o2c(connection_string, prefix=prefix)
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["case:concept:name", "org:resource"]
    )


def extract_ocel_sap_accounting(connection_string: str, prefix: str = '') -> OCEL:
    """
    Extracts an object-centric event log for the SAP Accounting process.

    Columns:
    - **ACTIVITY (ocel:activity)**: Activity performed in the Accounting process.
    - **TIMESTAMP (ocel:timestamp)**: Timestamp of the activity execution.

    Object Types:
    - **case:concept:name**: Identifier of the case.
    - **org:resource**: Resource involved in the activity.

    :param connection_string: ODBC connection string to the SAP database.
    :param prefix: Prefix for the tables (e.g., SAPSR3.).
    :rtype: ``OCEL``

    .. code-block:: python3
        import pm4py

        ocel = pm4py.connectors.extract_ocel_sap_accounting(
            'Driver={Oracle in instantclient_21_6};DBQ=127.0.0.3:1521/ZIB;UID=xx;PWD=yy'
        )
    """
    import pm4py
    dataframe = pm4py.connectors.extract_log_sap_accounting(connection_string, prefix=prefix)
    return pm4py.convert_log_to_ocel(
        dataframe,
        case_id_col="concept:name",
        timestamp_col="time:timestamp",
        object_attributes=["case:concept:name", "org:resource"]
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.evaluation.replay_fitness.variants import alignment_based, token_replay
from pm4py.algo.conformance import alignments
from pm4py.util import exec_utils
from pm4py.objects.petri_net.utils.check_soundness import check_easy_soundness_net_in_fin_marking
from enum import Enum
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
import pandas as pd


class Variants(Enum):
    ALIGNMENT_BASED = alignment_based
    TOKEN_BASED = token_replay


class Parameters(Enum):
    ALIGN_VARIANT = "align_variant"


ALIGNMENT_BASED = Variants.ALIGNMENT_BASED
TOKEN_BASED = Variants.TOKEN_BASED

VERSIONS = {ALIGNMENT_BASED, TOKEN_BASED}


def apply(log: Union[EventLog, pd.DataFrame], petri_net: PetriNet, initial_marking: Marking, final_marking: Marking, parameters: Optional[Dict[Union[str, Parameters], Any]] = None, variant=None, align_variant=None) -> Dict[str, Any]:
    """
    Apply fitness evaluation starting from an event log and a marked Petri net,
    by using one of the replay techniques provided by PM4Py

    Parameters
    -----------
    log
        Trace log object
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters related to the replay algorithm
    variant
        Chosen variant:
            - Variants.ALIGNMENT_BASED
            - Variants.TOKEN_BASED
    align_variant
        Alignments variant (for alignment-based replay)

    Returns
    ----------
    fitness_eval
        Fitness evaluation
    """
    if parameters is None:
        parameters = {}

    # execute the following part of code when the variant is not specified by the user
    if variant is None:
        if not (
                check_easy_soundness_net_in_fin_marking(petri_net, initial_marking,
                                                                              final_marking)):
            # in the case the net is not a easy sound workflow net, we must apply token-based replay
            variant = TOKEN_BASED
        else:
            # otherwise, use the align-etconformance approach (safer, in the case the model contains duplicates)
            variant = ALIGNMENT_BASED

    if variant == TOKEN_BASED:
        # execute the token-based replay variant
        return exec_utils.get_variant(variant).apply(log,
                                                     petri_net,
                                                     initial_marking, final_marking, parameters=parameters)
    else:
        # execute the alignments based variant, with the specification of the alignments variant
        if align_variant is None:
            align_variant = alignments.petri_net.algorithm.DEFAULT_VARIANT
        return exec_utils.get_variant(variant).apply(log,
                                                     petri_net,
                                                     initial_marking, final_marking, align_variant=align_variant,
                                                     parameters=parameters)


def evaluate(results, parameters=None, variant=TOKEN_BASED):
    """
    Evaluate replay results when the replay algorithm has already been applied

    Parameters
    -----------
    results
        Results of the replay algorithm
    parameters
        Possible parameters passed to the evaluation
    variant
        Indicates which evaluator is called

    Returns
    -----------
    fitness_eval
        Fitness evaluation
    """
    return exec_utils.get_variant(variant).evaluate(results, parameters=parameters)
